/*

  PARSEGDB.C - Garmin MapSource GDB file parser - works with MapSource v 6.3 and v 6.5 Beta

  Version 2.6

  Usage - parsegdb [-options] infile[.gdb] outfile[.gdb]

  Reads autogenerated routes from infile and writes them to outfile as tracks.

  Some of the information on the GDB format was provided by Ian Cowley and AnnoyedOne

  This program is derived from my observations of the data files produced by MapSource, and 
  is used at your own risk. No warranty is offered, expressed or implied.

  Copyright 2004 Simon Long
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "stdafx.h"

/* global store for name data */
char namebuf[30000];

/* globals to store position data */
unsigned long lat[10000]; 
unsigned long lon[10000];
unsigned long scount = 0L;

/* global to store number of waypoints created */
unsigned int new_points = 1;

/* global to store file version - 'k' = MS6.3, 'l' = MS6.5 Beta */
unsigned char version;

/* data lengths */
#define POS_LEN		(9)		/* position - lat/long/alt flag */
#define NUM_LEN		(4)		/* counter of number of points/links */
#define WP_HDR_BLK	(27)	/* waypoint header */
#define WP_TRL_BLK	(31)	/* waypoint trailer */
#define RT_HDR_BLK	(2)		/* route header */
#define RT_TRL_BLK	(2)		/* route trailer */
#define RP_HDX_BLK	(19)	/* additional waypoint header data in route */
#define RP_TRL_BLK	(8)		/* trailer bytes for route point in MS 6.5 */
#define TR_HDR_BLK  (5)		/* track header */
#define TR_TRL_BLK  (1)		/* track trailer */
#define TP_TRL_BLK	(3)		/* track point trailer */

/* size of waypoint, less waypoint name */
#define WP_POS_LEN	(WP_HDR_BLK + POS_LEN + WP_TRL_BLK)

/* size of track header and trailer, less track name */
#define TR_HDR_LEN	(TR_HDR_BLK + NUM_LEN + TR_TRL_BLK)	

/* size of track point */	
#define TR_POS_LEN	(POS_LEN + TP_TRL_BLK)

/* size of route header and trailer, less route name and name of final point */
#define RT_HDR_LEN	(RT_HDR_BLK + (POS_LEN * 2) + NUM_LEN + WP_HDR_BLK + RP_HDX_BLK + NUM_LEN + RT_TRL_BLK)

/* size of route point */
#define RT_POS_LEN	(WP_HDR_BLK + RP_HDX_BLK + NUM_LEN + (POS_LEN * 4) + 1)

/* header data blocks for waypoints, routes and tracks */
unsigned char wp_hdr[WP_HDR_BLK] = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF
};

unsigned char wp_trl[WP_TRL_BLK] = {
	0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x12, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00	
};

unsigned char rp_hdr_ext[RP_HDX_BLK] = {
	0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00
};

unsigned char rp_trl[RP_TRL_BLK] = {
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

unsigned char tr_hdr[TR_HDR_BLK] = {
	0x01, 0x02, 0x00, 0x00, 0x00
};

unsigned char tp_trl[TP_TRL_BLK] = {
	0x00, 0x00, 0x00
};

/* read a long from a file as 4 bytes, LSB first */
unsigned long read_long (FILE *fp)
{
	unsigned long val;
	unsigned char *cptr;
	int count;

	cptr = (unsigned char *) &val;
	for (count = 0; count < 4; count++)
		*cptr++ = fgetc (fp);
	return val;
}

/* read a null-terminated text string from a file */
unsigned int read_name (FILE *fp, char *name)
{
	unsigned int inchar, count = 0;

	while (inchar = fgetc (fp))
	{
		if (name != NULL) *name++ = inchar;
		count++;
	}
	if (name != NULL) *name = 0;
	count++;
	return count;
}

/* write a long to a file as 4 bytes, LSB first */
void write_long (unsigned long val, FILE *fp)
{
	unsigned char *cptr;
	int count;

	cptr = (unsigned char *) &val;
	for (count = 0; count < 4; count++)
		fputc (*cptr++, fp);
}

/* write a null-terminated text string to a file */
void write_name (char *name, FILE *fp)
{
	while (*name) fputc (*name++, fp);
	fputc (0x00, fp);
}

/* write a data block to a file */
void write_block (FILE *fp, unsigned char *ptr, int length)
{
	int count;

	for (count = 0; count < length; count++)
		fputc (*ptr++, fp);
}

/* write a position and a null altitude flag to a file */
void write_pos (unsigned long lat, unsigned long lon, FILE *fp)
{
	write_long (lat, fp);
	write_long (lon, fp);
	fputc (0x00, fp);
}

/* modified version of fputc - only writes if file pointer is non-NULL */
int tfputc (int val, FILE *fp)
{
	if (fp) fputc (val, fp);
	return val;
}

/* copy a number of bytes from the input file to the output file - if bytes = 0, copy until 0 found */
unsigned long copy_bytes (FILE *finp, FILE *foutp, unsigned long bytes)
{
	unsigned long count = 0;

	if (bytes)
	{
		/* loop through required number of bytes */
		for (count = 0; count < bytes; count++)
			tfputc (fgetc (finp), foutp);
	}
	else
	{
		/* loop until a zero is found on the input - for null-terminated strings */
		while (tfputc (fgetc (finp), foutp)) count++;
		count++;
	}
	return count;
}

/* copy a flagged set of bytes from the input file to the output file */
unsigned int copy_flagged (FILE *finp, FILE *foutp, unsigned int bytes)
{
	/* read the flag and value if present */
	if (tfputc (fgetc (finp), foutp))
	{
		copy_bytes (finp, foutp, bytes);
		return bytes + 1;
	}
	else return 1;
}

/* read the size and type of a chunk */
unsigned char read_chunk_header (FILE *fp, unsigned long *bytes)
{
	unsigned char type;

	/* read length of the chunk */
	*bytes = read_long (fp);

	/* read the chunk type */
	type = (unsigned char) fgetc (fp);

	return type;
}

/* write chunk header (size & type) data */
void write_chunk_header (FILE *fp, unsigned char type, unsigned long bytes)
{
	/* write the data length */
	write_long (bytes, fp);

	/* write the type */
	fputc (type, fp);
}

/* write a route latitude and longitude to the globals, ignoring duplicates */
void store_lat_long (unsigned long sa, unsigned long sb)
{
	if (scount == 10000L)
	{
		printf ("Route point limit reached - track truncated!\n");
		return;
	}

	if (scount == 0L)
	{
		lat[scount] = sa;
		lon[scount++] = sb;
	}
	else if (sa != lat[scount - 1] || sb != lon[scount - 1])
	{
		lat[scount] = sa;
		lon[scount++] = sb;
	}
}

/* calculates the maximum and minimum latitude and longitude in the specified range of the global table */
void min_max (unsigned int start, unsigned int end, unsigned long *min_lat, unsigned long *max_lat, unsigned long *min_long, unsigned long *max_long)
{
	unsigned int count;

	*min_lat = 0xFFFFFFFFL;
	*min_long = 0xFFFFFFFFL;
	*max_lat = 0;
	*max_long = 0;

	for (count = start; count <= end; count++)
	{
		if (lat[count] > *max_lat) *max_lat = lat[count];
		if (lat[count] < *min_lat) *min_lat = lat[count];
		if (lon[count] > *max_long) *max_long = lon[count];
		if (lon[count] < *min_long) *min_long = lon[count];
	}
}

/* read in a waypoint and modify the type bytes when writing it out */
void modify_waypoint (FILE *finp, FILE *foutp, unsigned long bytes, unsigned char mode)
{
	int type, inchar;

	/* get waypoint class */
	type = fgetc (finp);
	bytes--;

	/* change if 0x0A or 0x0C */
	if (type == 0x0A || type == 0x0C)
	{
		if (mode == 'R') 
			fputc (0x00, foutp);
		else 
			fputc (0x0C, foutp);
	}
	else fputc (type, foutp);

	/* skip 12 bytes */
	bytes -= copy_bytes (finp, foutp, 12);

	/* read the visible flag */
	inchar = fgetc (finp);
	bytes--;

	switch (type)
	{
		case 0x0C : if (inchar == 0x21) fputc (0xFF, foutp);
					else fputc (inchar, foutp);
					break;

		case 0x0A : if (inchar == 0x1F) fputc (0xFF, foutp);
					else fputc (inchar, foutp);
					break;

		default :	fputc (inchar, foutp);
					break;
	}

	/* copy the rest of the chunk unchanged */
	copy_bytes (finp, foutp, bytes);
}

/* read in a route and modify the type bytes for each point when writing it out */
void modify_route (FILE *finp, FILE *foutp, unsigned long bytes, unsigned char mode)
{
	unsigned long points, links;

	/* copy header */
	bytes -= copy_bytes (finp, foutp, 2);

	/* copy maximum and minimum positions and altitudes */
	bytes -= copy_bytes (finp, foutp, 8);
	bytes -= copy_flagged (finp, foutp, 8);
	bytes -= copy_bytes (finp, foutp, 8);
	bytes -= copy_flagged (finp, foutp, 8);

	/* read number of route links */
	points = read_long (finp);
	write_long (points, foutp);
	bytes -= 4;
	
	/* loop through points */
	while (points--)
	{
		/* copy the point name */
		bytes -= copy_bytes (finp, foutp, 0);

		/* modify the route header data */
		modify_waypoint (finp, foutp, 46, mode);
		bytes -= 46;

		/* read in number of sub points */
		links = read_long (finp);
		write_long (links, foutp);
		bytes -= 4;

		if (!links) break;

		/* pass through all sub points */
		while (links--)
		{
			/* copy position and altitude */
			bytes -= copy_bytes (finp, foutp, 8);
			bytes -= copy_flagged (finp, foutp, 8);
		}

		/* copy spacer and maximum and minimum position and altitude */
		bytes -= copy_bytes (finp, foutp, 9);
		bytes -= copy_flagged (finp, foutp, 8);
		bytes -= copy_bytes (finp, foutp, 8);
		bytes -= copy_flagged (finp, foutp, 8);

		/* in MapSource 6.5, there are an additional 8 bytes here - all 0xFF */
		if (version == 'l') bytes -= copy_bytes (finp, foutp, 8);
	}

	/* skip any remaining bytes */
	copy_bytes (finp, foutp, bytes);
}

/* route parsing routine - pulls out latitudes and longitudes and stores in global table */
unsigned int parse_route (FILE *fp, unsigned long bytes, char *name, unsigned char mode)
{
	unsigned long tlat, tlon, points, links;
	unsigned int first, namelen;

	/* clear the globals */
	scount = 0L;

	/* read in the route name */
	namelen = read_name (fp, name);
	bytes -= namelen;
		
	/* skip header */
	bytes -= copy_bytes (fp, NULL, 2);

	/* skip maximum and minimum positions and altitudes */
	bytes -= copy_bytes (fp, NULL, 8);
	bytes -= copy_flagged (fp, NULL, 8);
	bytes -= copy_bytes (fp, NULL, 8);
	bytes -= copy_flagged (fp, NULL, 8);

	/* read number of route links */
	points = read_long (fp);
	bytes -= 4;
	
	/* loop through points */
	while (points--)
	{
		/* skip point name */
		bytes -= read_name (fp, NULL);

		/* skip point header */
		bytes -= copy_bytes (fp, NULL, 46);

		/* read in number of sub points */
		links = read_long (fp);
		bytes -= 4;

		if (!links) break;

		first = 1;
		while (links--)
		{
			/* read the link position */
			tlat = read_long (fp);
			tlon = read_long (fp);
			bytes -= 8;

			/* skip the altitude */
			bytes -= copy_flagged (fp, NULL, 8);

			if (mode == 'T') store_lat_long (tlat, tlon);
			else
			{
				/* only store first and last link points */
				if (first || !links) store_lat_long (tlat, tlon);
				first = 0;
			}
		}

		/* skip spacer byte */
		bytes -= copy_bytes (fp, NULL, 1);

		/* skip maximum and minimum positions */
		bytes -= copy_bytes (fp, NULL, 8);
		bytes -= copy_flagged (fp, NULL, 8);
		bytes -= copy_bytes (fp, NULL, 8);
		bytes -= copy_flagged (fp, NULL, 8);

		/* in MapSource 6.5, there are an additional 8 bytes here - all 0xFF */
		if (version == 'l') bytes -= copy_bytes (fp, NULL, 8);
	}

	/* skip any remaining bytes */
	copy_bytes (fp, NULL, bytes);

	/* return length of the name */
	return namelen;
}

/* track parsing routine - pulls out latitudes and longitudes and stores in global table */
unsigned int parse_track (FILE *fp, unsigned long bytes, char *name)
{
	unsigned long tlat, tlon, points;
	unsigned int namelen;

	/* clear the globals */
	scount = 0L;

	/* read in the track name */
	namelen = read_name (fp, name);
	bytes -= namelen;
		
	/* skip header */
	bytes -= copy_bytes (fp, NULL, 5);

	/* read number of track points */
	points = read_long (fp);
	bytes -= 4;
	
	/* loop through points */
	while (points--)
	{
		/* read latitude and longitude */
		tlat = read_long (fp);
		tlon = read_long (fp);
		store_lat_long (tlat, tlon);
		bytes -= 8;

		/* skip altitude */
		bytes -= copy_flagged (fp, NULL, 8);

		/* skip time value */
		bytes -= copy_flagged (fp, NULL, 4);

		/* skip fillers */
		bytes -= copy_bytes (fp, NULL, 2);
	}

	/* skip any remaining bytes */
	copy_bytes (fp, NULL, bytes);

	/* return length of the name */
	return namelen;
}

/* writes a waypoint with the supplied latitude and longitude */
void write_waypoint (FILE *fp, char *name, unsigned long lat, unsigned long lon, int visible)
{
	/* write waypoint name */
	write_name (name, fp);

	/* write waypoint header */
	if (visible) write_block (fp, wp_hdr, WP_HDR_BLK);
	else 
	{
		/* change class byte to make via point */
		fputc (0x0C, fp);
		write_block (fp, wp_hdr + 1, WP_HDR_BLK - 1);
	}

	/* write latitude and longitude and altitude */
	write_pos (lat, lon, fp);

	/* write waypoint trailer bytes */
	if (visible) write_block (fp, wp_trl, WP_TRL_BLK);
	else
	{
		/* change show name byte */
		write_block (fp, wp_trl, 2);
		fputc (0x00, fp);
		write_block (fp, wp_trl + 3, WP_TRL_BLK - 3);
	}
}

/* route writing routine - writes latitudes and longitudes stored in global table */
void write_route (FILE *fp, char *name, int visible)
{
	unsigned long min_lat, max_lat, min_long, max_long;
	unsigned int point;

	/* write route name */
	write_name (name, fp);

	/* write header bytes */
	fputc (0x01, fp);
	fputc (0x00, fp);

	/* write route area - maximum and minimum latitude and longitude */
	min_max (0, scount - 1, &min_lat, &max_lat, &min_long, &max_long);
	write_pos (max_lat, max_long, fp);
	write_pos (min_lat, min_long, fp);

	/* write number of route links */
	write_long (scount, fp);
	
	/* loop through links */
	for (point = 0; point < scount; point++)
	{
		/* prepare point name */
		sprintf (namebuf, "%03d", point + new_points);

		/* write point name */
		write_name (namebuf, fp);

		/* write waypoint header */
		if (visible) write_block (fp, wp_hdr, WP_HDR_BLK);
		else 
		{
			/* change class byte to make via point */
			fputc (0x0C, fp);
			write_block (fp, wp_hdr + 1, WP_HDR_BLK - 1);
		}

		/* write additional route point header */
		write_block (fp, rp_hdr_ext, RP_HDX_BLK);

		if (point == scount - 1)
		{
			/* write zero link points to terminate */
			write_long (0, fp);

			/* write trailer bytes - include extra 8 bytes with 6.5 Beta */
			fputc (0x01, fp);
			if (version == 'l') write_block (fp, rp_trl, RP_TRL_BLK);
			fputc (0x00, fp);
		}
		else
		{
			/* write number of link points - always 2 in this case */
			write_long (2, fp);

			/* write 2 link points */
			write_pos (lat[point], lon[point], fp);
			write_pos (lat[point + 1], lon[point + 1], fp);

			/* separator byte */
			fputc (0x00, fp);

			/* write area marks - maximum and minimum latitude and longitude of route segment */
			min_max (point, point + 1, &min_lat, &max_lat, &min_long, &max_long);
			write_pos (max_lat, max_long, fp);
			write_pos (min_lat, min_long, fp);

			/* in MapSource 6.5, there are an additional 8 bytes here - all 0xFF */
			if (version == 'l') write_block (fp, rp_trl, RP_TRL_BLK); 
		}
	}
}

/* writes a track to the output file based on contents of the globals */
void write_track (FILE *fp, char *name)
{
	unsigned long count;

	/* write track name */
	write_name (name, fp);

	/* track header bytes */
	write_block (fp, tr_hdr, TR_HDR_BLK);

	/* write the number of points */
	write_long (scount, fp);

	/* loop through points */
	for (count = 0; count < scount; count++)
	{
		/* write position */
		write_pos (lat[count], lon[count], fp);

		/* write point trailer bytes */
		write_block (fp, tp_trl, TP_TRL_BLK);
	}

	/* write trailing zero */
	fputc (0x00, fp);
}

/* read in and process a single data chunk - return chunk type */
unsigned char process_chunk (FILE *finp, FILE *foutp, unsigned char mode)
{
	unsigned char type;
	unsigned int namelen, point;
	unsigned long chunk_length;
	char wpname[4];

	type = read_chunk_header (finp, &chunk_length);
	printf ("%c %d\n", type, chunk_length);

	/* parse the chunk */
	switch (type)
	{
		case 'T' :	/* track */
					if (mode == 'V' || mode == 'W')
					{
						/* parse track into global table */
						namelen = parse_track (finp, chunk_length, namebuf);

						/* check that no more than 999 waypoints will be created */
						if (scount + new_points > 1000)
						{
							printf ("Too many waypoints needed to convert - track skipped\n");
						}
						else
						{
							/* write global table as waypoints */
							for (point = 0; point < scount; point++)
							{
								/* write header for waypoint, assuming 3 character (+ null) name */
								write_chunk_header (foutp, 'W', (4 + WP_POS_LEN));

								/* generate name as three digit number from 001 upwards */
								sprintf (wpname, "%03d", point + new_points);

								/* write the waypoint */
								write_waypoint (foutp, wpname, lat[point], lon[point], mode == 'V' ? 1 : 0);
							}

							/* write the route header, assuming 3 character (+ null) names */
							if (version == 'k')
								write_chunk_header (foutp, 'R', (namelen + RT_HDR_LEN + 4 + ((scount - 1) * (RT_POS_LEN + 4))));
							else
								write_chunk_header (foutp, 'R', (namelen + RT_HDR_LEN + 4 + RP_TRL_BLK + ((scount - 1) * (RT_POS_LEN + 4 + RP_TRL_BLK))));


							/* write the route */
							write_route (foutp, namebuf, mode == 'V' ? 1 : 0);

							/* update the global */
							new_points += scount;
						}
					}
					else
					{
						/* just copy otherwise */
						write_chunk_header (foutp, type, chunk_length);
						copy_bytes (finp, foutp, chunk_length); 
					}
					break;

		case 'R' :	/* route */
					if (mode == 'R' || mode == 'S')
					{
						/* change type bytes if creating a route */
						write_chunk_header (foutp, type, chunk_length);
						chunk_length -= copy_bytes (finp, foutp, 0);
						modify_route (finp, foutp, chunk_length, mode);
					}
					else if (mode == 'T' || mode == 'U')
					{
						/* parse route into global table */
						namelen = parse_route (finp, chunk_length, namebuf, mode);

						/* write track from globals */
						write_chunk_header (foutp, 'T', (namelen + TR_HDR_LEN + (scount * TR_POS_LEN)));
						write_track (foutp, namebuf);
					}
					else
					{
						/* just copy otherwise */
						write_chunk_header (foutp, type, chunk_length);
						copy_bytes (finp, foutp, chunk_length); 
					}
					break;

		case 'W' :	/* waypoint */
					write_chunk_header (foutp, type, chunk_length);
					if (mode == 'R' || mode == 'S')
					{
						/* change type bytes if creating a route */
						chunk_length -= copy_bytes (finp, foutp, 0);
						modify_waypoint (finp, foutp, chunk_length, mode);
					}
					else
					{
						/* just copy otherwise */
						copy_bytes (finp, foutp, chunk_length); 
					}
					break;

		default :	/* other chunk type - copy across unchanged */
					write_chunk_header (foutp, type, chunk_length);
					copy_bytes (finp, foutp, chunk_length); 
					break;
	}

	return type;
}

/* process GDB header and check valid */
int copy_header (FILE *finp, FILE *foutp)
{
	/* MsRcf */
	read_name (finp, namebuf);
	if (strcmp (namebuf, "MsRcf")) return 0;
	write_name (namebuf, foutp);

	/* header data */
	copy_bytes (finp, foutp, 5);

	/* read and check MapSource version */
	version = fgetc (finp);
	fputc (version, foutp);
	if (version != 'k' && version != 'l') return 0;

	/* rest of header */
	copy_bytes (finp, foutp, 8);

	/* 'SQA' - 'hales' in MS 6.5 Beta */
	copy_bytes (finp, foutp, 0);

	/* date */
	copy_bytes (finp, foutp, 0);

	/* time */
	copy_bytes (finp, foutp, 0);

	/* MapSource */
	read_name (finp, namebuf);
	if (strcmp (namebuf, "MapSource")) return 0;
	write_name (namebuf, foutp);

	return 1;
}

#ifndef _WINDOWS
/* help text */
void usage (void)
{
	printf ("Usage = parsegdb [-R|-S|-T|-U|-V|-W] infile[.gdb] outfile[.gdb]\n");
	printf ("  -R = Convert routes to route with user vias\n");
	printf ("  -S = Convert routes to route with hidden vias\n");
	printf ("  -T = Convert routes to track with all points (default)\n");
	printf ("  -U = Convert routes to track with via points only\n");
	printf ("  -V = Convert tracks to route with user vias\n");
	printf ("  -W = Convert tracks to route with hidden vias\n");
	exit (0);
}

/* process the command line parameters */
unsigned char parse_command_line (int argc, char *argv[], FILE **finp, FILE **foutp)
{
	unsigned char mode = 'T';	/* default */
	unsigned char param = 1;

	/* need at least 3 parameters including exe name */
	if (argc < 3) usage ();

	/* is the first parameter an option? */
	if (*argv[1] == '-')
	{
		if (argv[1][1] == 'R' || argv[1][1] == 'r') mode = 'R'; 
		if (argv[1][1] == 'S' || argv[1][1] == 's') mode = 'S'; 
		if (argv[1][1] == 'U' || argv[1][1] == 'u') mode = 'U'; 
		if (argv[1][1] == 'V' || argv[1][1] == 'v') mode = 'V'; 
		if (argv[1][1] == 'W' || argv[1][1] == 'w') mode = 'W'; 
		param++;
	}

	/* are there still 2 file names? */
	if (argc != param + 2) usage ();

	/* open infile, adding extension if needed */
	strcpy (namebuf, argv[param]);
	if (!strchr (namebuf, '.')) strcat (namebuf, ".gdb");
	*finp = fopen (namebuf, "rb");
	if (*finp == NULL) 
	{
		printf ("Can't open input file\n");
		exit (0);
	}

	/* open outfile, adding extension if needed */
	strcpy (namebuf, argv[param + 1]);
	if (!strchr (namebuf, '.')) strcat (namebuf, ".gdb");
	*foutp = fopen (namebuf, "wb");
	if (*foutp == NULL) 
	{
		printf ("Can't open output file\n");
		exit (0);
	}

	return mode;
}

/* main routine */
void main (int argc, char *argv[])
{
	FILE *finp, *foutp;
	unsigned char mode;

	/* parse command line */
	mode = parse_command_line (argc, argv, &finp, &foutp);

	/* copy the file header */
	if (copy_header (finp, foutp))
	{
		/* scan chunks until chunk type 'V' is found - always seems to be last in file */
		while (process_chunk (finp, foutp, mode) != 'V');
	}
	else printf ("Input file not in GDB format\n");

	/* close files */
	fclose (foutp);
	fclose (finp);
}
#endif
/* End of file */
